@@ -2,9 +2,9 @@
 #
 #  build-initrd.sh
 #
-#  Written by Jari Ruusu, May 8 2004
+#  Written by Jari Ruusu, June 17 2005
 #
-#  Copyright 2001,2002,2003,2004 by Jari Ruusu.
+#  Copyright 2001-2005 by Jari Ruusu.
 #  Redistribution of this file is permitted under the GNU Public License.
 #
 #  Changes by Hauke Johannknecht <ash@ash.de> 11/2001
@@ -156,9 +156,13 @@ LOOPMODPARAMS=""
 LOADNATIONALKEYB=0
 
 # Initial delay in seconds before /linuxrc attempts to mount /boot
-# partition. Slow devices (USB dongles) may need some delay.
+# partition. Slow devices (USB-sticks) may need some delay.
 INITIALDELAY=0
 
+# Delay in seconds before /linuxrc attempts to mount partition containing
+# external gpg key files. Slow devices (USB-sticks) may need some delay.
+MOUNTDELAY=0
+
 # 1 = prompt for BOOT-TOOLS media and ENTER press before mounting /boot
 # 0 = normal case, don't prompt
 TOOLSPROMPT=0
@@ -487,6 +491,14 @@ void _start()
 #endif
 
 #if ${EXTERNALGPGFILES}
+#if ${MOUNTDELAY}
+    {
+        struct timespec req;
+        req.tv_sec = ${MOUNTDELAY};
+        req.tv_nsec = 0;
+        nanosleep(&req, 0);
+    }
+#endif
     if(mount("${EXTERNALGPGDEV}", "/mnt", "${EXTERNALGPGTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
         wrStr("Mounting ${EXTERNALGPGDEV} containing gpg key files failed.\n");
         goto fail5;
@@ -547,12 +559,15 @@ void _start()
     }
 #endif
 
-    close(0);
-    close(1);
-    close(2);
-    open("dev/console", O_RDWR, 0);
-    dup(0);
-    dup(0);
+    x = open("dev/console", O_RDWR, 0);
+    if(x == -1) {
+        wrStr("Opening /dev/console on encrypted root failed\n");
+        goto fail1;
+    }
+    dup2(x, 0);
+    dup2(x, 1);
+    dup2(x, 2);
+    close(x);
 
     if(chroot(".")) {
         wrStr("chroot() to new root failed\n");
@@ -614,9 +629,7 @@ if [ ${EXTERNALGPGFILES} == 1 ] ; then
 fi
 
 dd if=/dev/zero of=tmp-i-$$ bs=1024 count=${y}
-/sbin/losetup /dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX} tmp-i-$$
-/sbin/mkfs -t minix -i 32 /dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
-/sbin/losetup -d /dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
+/sbin/mkfs -t minix -i 32 tmp-i-$$
 mkdir tmp-d-$$
 mount -t minix tmp-i-$$ tmp-d-$$ -o loop=/dev/loop${DEVFSSLASH2}${TEMPLOOPINDEX}
 cd tmp-d-$$
@@ -740,7 +753,7 @@ fi
 for x in ${z} ; do
     echo Copying ${SOURCEROOT}${x} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
     cp -p ${SOURCEROOT}${x} ${DESTINATIONROOT}${DESTINATIONPREFIX}
-    y=`ldd ${SOURCEROOT}${x} | perl -ne 'if(/ => ([^ ]*) /){print "$1\n"}'`
+    y=`ldd ${SOURCEROOT}${x} | perl -ne 'if(/([^ ]*) \(0x/){print "$1\n"}'`
     for a in ${y} ; do
         echo Copying ${SOURCEROOT}${a} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
         cp -p ${SOURCEROOT}${a} ${DESTINATIONROOT}${DESTINATIONPREFIX}
