@@ -1,9 +1,9 @@
 /*
  *  glue.c
  *
- *  Written by Jari Ruusu, August 14 2004
+ *  Written by Jari Ruusu, April 26 2005
  *
- *  Copyright 2001,2002,2003,2004 by Jari Ruusu.
+ *  Copyright 2001-2005 by Jari Ruusu.
  *  Redistribution of this file is permitted under the GNU Public License.
  */
 
@@ -56,6 +56,9 @@ typedef int TransferSector_t;
 #if !defined(LOOP_MULTI_KEY_SETUP)
 # define LOOP_MULTI_KEY_SETUP 0x4C4D
 #endif
+#if !defined(LOOP_MULTI_KEY_SETUP_V3)
+# define LOOP_MULTI_KEY_SETUP_V3 0x4C4E
+#endif
 
 #if CONFIG_BLK_DEV_LOOP_KEYSCRUB
 # define KEY_ALLOC_COUNT  128
@@ -67,10 +70,14 @@ typedef struct {
     aes_context *keyPtr[KEY_ALLOC_COUNT];
     unsigned    keyMask;
 #if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+    u_int32_t   *partialMD5;
+    u_int32_t   partialMD5buf[8];
     rwlock_t    rwlock;
     unsigned    reversed;
     unsigned    blocked;
     struct timer_list timer;
+#else
+    u_int32_t   partialMD5[4];
 #endif
 } AESmultiKey;
 
@@ -96,7 +103,17 @@ static void keyScrubWork(AESmultiKey *m)
             y--;
         }
     }
-    m->reversed ^= 1;
+
+    x = m->reversed;    /* x is 0 or 4 */
+    m->reversed ^= 4;
+    y = m->reversed;    /* y is 4 or 0 */
+    p = &m->partialMD5buf[x];
+    memcpy(&m->partialMD5buf[y], p, 16);
+    m->partialMD5 = &m->partialMD5buf[y];
+    p[0] ^= 0xFFFFFFFF;
+    p[1] ^= 0xFFFFFFFF;
+    p[2] ^= 0xFFFFFFFF;
+    p[3] ^= 0xFFFFFFFF;
 
     /* try to flush dirty cache data to RAM */
 #if defined(CONFIG_X86_64) || (defined(CONFIG_X86) && !defined(CONFIG_M386) && !defined(CONFIG_CPU_386))
@@ -114,11 +131,12 @@ static void keyScrubThreadFn(AESmultiKey *m)
     write_unlock(&m->rwlock);
 }
 
+static void keyScrubTimerFn(unsigned long);
+
 static void keyScrubTimerInit(struct loop_device *lo)
 {
     AESmultiKey     *m;
     unsigned long   expire;
-    static void keyScrubTimerFn(unsigned long);
 
     m = (AESmultiKey *)lo->key_data;
     expire = jiffies + HZ;
@@ -152,6 +170,7 @@ static AESmultiKey *allocMultiKey(void)
     if(!m) return 0;
     memset(m, 0, sizeof(AESmultiKey));
 #if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+    m->partialMD5 = &m->partialMD5buf[0];
     rwlock_init(&m->rwlock);
     init_timer(&m->timer);
     again:
@@ -204,17 +223,18 @@ static void clearAndFreeMultiKey(AESmultiKey *m)
         x += n;
     }
 
+    memset(m, 0, sizeof(AESmultiKey));
     kfree(m);
 }
 
-static int multiKeySetup(struct loop_device *lo, unsigned char *k)
+static int multiKeySetup(struct loop_device *lo, unsigned char *k, int version3)
 {
     AESmultiKey *m;
     aes_context *a;
     int x, y, n, err = 0;
     union {
-        u_int32_t     w[8]; /* needed for 4 byte alignment for b[] */
-        unsigned char b[32];
+        u_int32_t     w[16];
+        unsigned char b[64];
     } un;
 
 #if LINUX_VERSION_CODE >= 0x20200
@@ -265,6 +285,28 @@ static int multiKeySetup(struct loop_device *lo, unsigned char *k)
         k += 32;
         x++;
     }
+
+    m->partialMD5[0] = 0x67452301;
+    m->partialMD5[1] = 0xefcdab89;
+    m->partialMD5[2] = 0x98badcfe;
+    m->partialMD5[3] = 0x10325476;
+    if(version3) {
+        /* only first 128 bits of iv-key is used */
+        if(copy_from_user(&un.b[0], k, 16)) {
+            err = -EFAULT;
+            goto error_out;
+        }
+#if defined(__BIG_ENDIAN)
+        un.w[0] = cpu_to_le32(un.w[0]);
+        un.w[1] = cpu_to_le32(un.w[1]);
+        un.w[2] = cpu_to_le32(un.w[2]);
+        un.w[3] = cpu_to_le32(un.w[3]);
+#endif
+        memset(&un.b[16], 0, 48);
+        md5_transform_CPUbyteorder(&m->partialMD5[0], &un.w[0]);
+        lo->lo_flags |= 0x080000;  /* multi-key-v3 (info exported to user space) */
+    }
+
     m->keyMask = 0x3F;          /* range 0...63 */
     lo->lo_flags |= 0x100000;   /* multi-key (info exported to user space) */
     memset(&un.b[0], 0, 32);
@@ -290,7 +332,7 @@ void loop_compute_sector_iv(TransferSector_t devSect, u_int32_t *ivout)
     }
 }
 
-void loop_compute_md5_iv(TransferSector_t devSect, u_int32_t *ivout, u_int32_t *data)
+void loop_compute_md5_iv_v3(TransferSector_t devSect, u_int32_t *ivout, u_int32_t *data)
 {
     int         x;
 #if defined(__BIG_ENDIAN)
@@ -298,11 +340,6 @@ void loop_compute_md5_iv(TransferSector_t devSect, u_int32_t *ivout, u_int32_t *
 #endif
     u_int32_t   buf[16];
 
-    ivout[0] = 0x67452301;
-    ivout[1] = 0xefcdab89;
-    ivout[2] = 0x98badcfe;
-    ivout[3] = 0x10325476;
-
 #if defined(__BIG_ENDIAN)
     y = 7;
     e = 16;
@@ -321,6 +358,8 @@ void loop_compute_md5_iv(TransferSector_t devSect, u_int32_t *ivout, u_int32_t *
                 buf[13] = 0x80000000;
             }
             /* 4024 bits == 31 * 128 bit plaintext blocks + 56 bits of sector number */
+            /* For version 3 on-disk format this really should be 4536 bits, but can't be */
+            /* changed without breaking compatibility. V3 uses MD5-with-wrong-length IV */
             buf[14] = 4024;
             buf[15] = 0;
         }
@@ -358,12 +397,31 @@ void loop_compute_md5_iv(TransferSector_t devSect, u_int32_t *ivout, u_int32_t *
         buf[13] = 0x80000000;
     }
     /* 4024 bits == 31 * 128 bit plaintext blocks + 56 bits of sector number */
+    /* For version 3 on-disk format this really should be 4536 bits, but can't be */
+    /* changed without breaking compatibility. V3 uses MD5-with-wrong-length IV */
     buf[14] = 4024;
     buf[15] = 0;
     md5_transform_CPUbyteorder(&ivout[0], &buf[0]);
 #endif
 }
 
+/* this function exists for compatibility with old external cipher modules */
+void loop_compute_md5_iv(TransferSector_t devSect, u_int32_t *ivout, u_int32_t *data)
+{
+    ivout[0] = 0x67452301;
+    ivout[1] = 0xefcdab89;
+    ivout[2] = 0x98badcfe;
+    ivout[3] = 0x10325476;
+    loop_compute_md5_iv_v3(devSect, ivout, data);
+}
+
+/* Some external modules do not know if md5_transform_CPUbyteorder() */
+/* is asmlinkage or not, so here is C language wrapper for them. */
+void md5_transform_CPUbyteorder_C(u_int32_t *hash, u_int32_t const *in)
+{
+    md5_transform_CPUbyteorder(hash, in);
+}
+
 int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
           char *loop_buf, int size, TransferSector_t devSect)
 {
@@ -402,7 +460,8 @@ int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
                 if(y && !x) {
                     raw_buf -= 496;
                     loop_buf -= 496;
-                    loop_compute_md5_iv(devSect, &iv[4], (u_int32_t *)(&loop_buf[16]));
+                    memcpy(&iv[4], &m->partialMD5[0], 16);
+                    loop_compute_md5_iv_v3(devSect, &iv[4], (u_int32_t *)(&loop_buf[16]));
                 } else {
                     raw_buf += 16;
                     loop_buf += 16;
@@ -454,7 +513,8 @@ int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
 # define TMP_RAW_BUF raw_buf
 #endif
                 memcpy(TMP_RAW_BUF, loop_buf, 512);
-                loop_compute_md5_iv(devSect, &iv[0], (u_int32_t *)(&TMP_RAW_BUF[16]));
+                memcpy(&iv[0], &m->partialMD5[0], 16);
+                loop_compute_md5_iv_v3(devSect, &iv[0], (u_int32_t *)(&TMP_RAW_BUF[16]));
                 x = 15;
                 do {
                     iv[0] ^= *((u_int32_t *)(&TMP_RAW_BUF[ 0]));
@@ -555,7 +615,10 @@ int handleIoctl_aes(struct loop_device *lo, int cmd, unsigned long arg)
 
     switch (cmd) {
     case LOOP_MULTI_KEY_SETUP:
-        err = multiKeySetup(lo, (unsigned char *)arg);
+        err = multiKeySetup(lo, (unsigned char *)arg, 0);
+        break;
+    case LOOP_MULTI_KEY_SETUP_V3:
+        err = multiKeySetup(lo, (unsigned char *)arg, 1);
         break;
     default:
         err = -EINVAL;
